// $Id: command.c,v 1.2 2003/11/06 04:55:19 ensc Exp $    --*- c -*--

// Copyright (C) 2003 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; version 2 of the License.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "command.h"
#include "djinniprocess.h"
#include "wrappers.h"
#include "wish.h"
#include "exithandler.h"
#include "util.h"
#include "util-io.h"
#include "logging.h"

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/wait.h>

#include "command-log.hc"

bool
Command_init(struct Command *self, struct DjinniProcess *proc, int fd)
{
  self->fd    = fd;
  self->proc_ = proc;
  
  if (!Parameters_init(&self->params)) {
    Xclose(self->fd);
    return false;
  }

  return true;
}

void
Command_detach(struct Command *self)
{
  XXclose(&self->fd);
}

void
Command_delete(struct Command *self)
{
  Command_detach(self);
  Parameters_delete(&self->params);
}

static struct DjinniProcess *
createNewDjinnie(struct Command *self)
{
  if (!DjinniProcess_chrootVserver(self->proc_)) {
    perror("DjinniProcess_chrootVserver()");
    exit(1);
  }
  
  Echdir(self->params.vserver_vdir);
  
  int		vserver_fd = Eopen(".", O_RDONLY, 0);

  if (!DjinniProcess_chdirState(self->proc_)) {
    perror("restoring djinni-statedir");
    exit(1);
  }

  char		new_cfg_dir[sizeof("../.cfg/") + strlen(self->params.argv[0])];

  strcpy(new_cfg_dir, "../.cfg/");
  strcat(new_cfg_dir, self->params.argv[0]);
  
  return DjinniProcess_new(vserver_fd,
			   new_cfg_dir,
			   self->params.vserver_name,
			   self->proc_, -1);
}

static bool
executeRun(struct Command *self)
{
  if (access(".run", X_OK)!=0) return true;

  pid_t		pid = fork();
  switch (pid) {
    case -1	:
      Logging_timestamp(2); perror("fork()");
      Wish_sendResult(self->fd, 255, "Internal error: failed to create run-process", 0);
      return false;

    case 0	:
    {
      ExitHandler_clear();
      Command_detach(self);

      char const	*cfg_dir_abs  = get_current_dir_name();
      size_t		l1            = strlen(cfg_dir_abs);
      char		run_file_abs[l1 + sizeof("/.run")];

      if (self->params.vserver_vdir) Esetenv  ("DJINNI_VSERVERDIR",
					       self->params.vserver_vdir, 1);
      else                           Eunsetenv("DJINNI_VSERERDIR");

      if (self->params.vserver_name) Esetenv  ("DJINNI_VSERVERNAME",
					       self->params.vserver_name, 1);
      else                           Eunsetenv("DJINNI_VSERVERNAME");
      
      strcpy(run_file_abs,    cfg_dir_abs);
      strcpy(run_file_abs+l1, "/.run");
      free(const_cast(char *)(cfg_dir_abs));

      DjinniProcess_chdirVserver(self->proc_);

      Logging_timestamp(1); WRITE_MSG(1, "starting external process\n");
      Eexecv(run_file_abs, self->params.argv);
      exit(1);
    }

    default	:
    {
      int		status;
      int		code;
      while (wait4(pid, &status,0,0)==-1) {}
      logProcessResult(pid, status);

      if (WIFEXITED(status)) code=WEXITSTATUS(status);
      else                   code=255;

      if (code!=0) {
	Wish_sendResult(self->fd, code, "FAIL: command failed",0);
	return false;
      }

      return true;
    }
  }
}

static bool
spawnNewDjinnie(struct Command *self)
{
  if (!access(".new", F_OK)==0) return true;
    

  int		sync_fd[2];
  if (pipe(sync_fd)==-1) {
    Logging_timestamp(2); perror("pipe()");
    Wish_sendResult(self->fd, 255, "Internal error: failed to create sync-pipe", 0);
    return false;
  }
  
  pid_t		pid = fork();
  switch (pid) {
    case -1	:
      Logging_timestamp(2); perror("fork()");
      Wish_sendResult(self->fd, 255, "Internal error: failed to create new djinni-process", 0);
      return false;

    case 0	:
    {
      struct DjinniProcess *	djinni;
      ExitHandler_clear();

      Logging_newProcess(self->params.vserver_name);
      
      close(sync_fd[0]);
      djinni = createNewDjinnie(self);
      DjinniProcess_detach(self->proc_);
      Command_detach(self);
      close(sync_fd[1]);
      exit(DjinniProcess_run(djinni));
    }

    default	:
    {
      char		tmp;
      int		status;

      close(sync_fd[1]);
      TEMP_FAILURE_RETRY(read(sync_fd[0], &tmp, sizeof tmp));
      close(sync_fd[0]);
      
      if (wait4(pid, &status, WNOHANG,0)!=0) {
	Logging_timestamp(2); WRITE_MSG(2, "failed to spawn new djinni process\n");
	logProcessResult(pid, status);
	
	Wish_sendResult(self->fd, 255, "Failed to spawn new djinni process",0);
	return false;
      }
    }
  }

  return true;
}

static void
executeExternal(struct Command *self)
{
  if (!DjinniProcess_chdirCfg(self->proc_) ||
      chdir(self->params.argv[0])==-1) {
    Logging_timestamp(2); WRITE_MSG(2, "failed to enter cfg-directory\n");
    Wish_sendResult(self->fd, 255, "Internal error: failed to enter cfg-directory", 0);
    return;
  }

  executeRun(self) && spawnNewDjinnie(self) &&
    Wish_sendResult(self->fd, 0, "OK", 0);
}

static bool
executeInternal(struct Command *self)
{
  if (self->params.argc==0) return false;

  char const	*argv0 = self->params.argv[0];
  bool		result = true;

  if      (strcmp(argv0, "ping")==0)
    Wish_sendResult(self->fd, 0, "pong", 0);
  else if (strcmp(argv0, "quit")==0) {
    Wish_sendResult(self->fd, 0, "", 0);
    DjinniProcess_doQuit(self->proc_);
  }
  else result = false;

  return result;
}

static bool
checkParams(struct Command *self)
{
  unsigned int		code;
  char const		*reason;
  
  if (!Parameters_evaluate(&self->params, self->proc_, &code, &reason)) {
    logCheckResult(code, reason);
    Wish_sendResult(self->fd, code, reason, 0);
    return false;
  }

  return true;
}

void
Command_execute(struct Command *self)
{
  if (!DjinniProcess_chdirState(self->proc_)) {
    Logging_timestamp(2); perror("DjinniProcess_chdirState()");
    Wish_sendResult(self->fd, 255, "Internal error: failed to enter state-directory", 0);
    return;
  }

  logCommand(&self->params);
  if (!executeInternal(self) && checkParams(self))
    executeExternal(self);
}
